use std::{
    sync::{
        mpsc::{channel, Receiver, Sender, TryRecvError},
        Arc, Mutex,
    },
    thread,
};

use crate::{
    context::{RWKVContext, RWKVInvocation},
    sampling::sample_logits,
    types::{
        EmbeddingResult, InferenceResult, InferenceResultType, InferenceToken, RWKVCommand,
        TokenizeResult, TokenizeResultType,
    },
};

#[derive(Clone)]
pub struct RWKVChannel {
    command_sender: Sender<RWKVCommand>,
    command_receiver: Arc<Mutex<Receiver<RWKVCommand>>>,
}

#[derive(Clone)]
pub struct RWKVInternal {
    context: RWKVContext,
}

impl RWKVInternal {
    pub fn tokenize(&self, input: &str, sender: &Sender<TokenizeResult>) {
        let tokenizer = &self.context.tokenizer;
        let tokens_result = tokenizer.encode(input, false).map(Some).unwrap_or(None);
        if let Some(result) = tokens_result {
            let tokens = result.get_ids().to_vec();
            sender
                .send(TokenizeResult {
                    r#type: TokenizeResultType::Data,
                    data: tokens.iter().map(|x| *x as i32).collect(),
                })
                .unwrap();
        } else {
            sender
                .send(TokenizeResult {
                    r#type: TokenizeResultType::Error,
                    data: vec![],
                })
                .unwrap();
        }
    }

    /* pub fn embedding(&self, input: &LlamaInvocation, sender: &Sender<EmbeddingResult>) {
        let context_params_c = LlamaContextParams::or_default(&self.context_params);
        let input_ctx = &self.context;
        let embd_inp = tokenize(
            input_ctx,
            input.prompt.as_str(),
            context_params_c.n_ctx as usize,
            true,
        )
        .unwrap();

        // let end_text = "\n";
        // let end_token =
        //     tokenize(input_ctx, end_text, context_params_c.n_ctx as usize, false).unwrap();

        input_ctx
            .llama_eval(embd_inp.as_slice(), embd_inp.len() as i32, 0, input)
            .unwrap();

        let embeddings = input_ctx.llama_get_embeddings();

        if let Ok(embeddings) = embeddings {
            sender
                .send(EmbeddingResult {
                    r#type: EmbeddingResultType::Data,
                    data: embeddings.iter().map(|&x| x as f64).collect(),
                })
                .unwrap();
        } else {
            sender
                .send(EmbeddingResult {
                    r#type: EmbeddingResultType::Error,
                    data: vec![],
                })
                .unwrap();
        }
    } */

    pub fn inference(&mut self, input: &RWKVInvocation, sender: &Sender<InferenceResult>) {
        let end_token = input.end_token.unwrap_or(0) as usize;
        let temp = input.temp as f32;
        let top_p = input.top_p as f32;
        let seed = input.seed.map(|x| x as u64);

        let context = &mut self.context;
        let tokenizer = &context.tokenizer;
        let prompt = &input.prompt;
        let binding = tokenizer.encode(prompt.as_str(), false).unwrap();
        let tokens = binding.get_ids().iter().map(|x| *x as i32).collect::<Vec<i32>>();

        let mut session = context.create_new_session();

        session.process_tokens(&tokens);

        for _i in 0..input.max_predict_length {
            let logits = session.logits.as_mut();
            let token = sample_logits(logits, temp, top_p, &seed);

            if token >= 50276 || token == end_token {
                sender
                    .send(InferenceResult {
                        r#type: InferenceResultType::Data,
                        message: None,
                        data: Some(InferenceToken {
                            token: "\n\n<end>\n".to_string(),
                            completed: true,
                        }),
                    })
                    .unwrap();
                return;
            }

            let decoded = context.rwkv_token_to_str(&(token as i32)).unwrap();

            sender
                .send(InferenceResult {
                    r#type: InferenceResultType::Data,
                    message: None,
                    data: Some(InferenceToken {
                        token: decoded,
                        completed: false,
                    }),
                })
                .unwrap();

            session.process_tokens(&[token.try_into().unwrap()]);
        }
    }
}

impl RWKVChannel {
    pub fn new(
        model_path: String,
        tokenizer_path: String,
        n_threads: u32,
        load_result_sender: Sender<bool>,
        enable_logger: bool,
    ) -> Arc<Self> {
        let (command_sender, command_receiver) = channel::<RWKVCommand>();

        let channel = RWKVChannel {
            command_receiver: Arc::new(Mutex::new(command_receiver)),
            command_sender,
        };

        channel.spawn(
            model_path,
            tokenizer_path,
            n_threads,
            load_result_sender,
            enable_logger,
        );

        Arc::new(channel)
    }

    pub fn tokenize(&self, input: String, sender: Sender<TokenizeResult>) {
        self.command_sender
            .send(RWKVCommand::Tokenize(input, sender))
            .unwrap();
    }

    pub fn embedding(&self, params: RWKVInvocation, sender: Sender<EmbeddingResult>) {
        self.command_sender
            .send(RWKVCommand::Embedding(params, sender))
            .unwrap();
    }

    pub fn inference(&self, params: RWKVInvocation, sender: Sender<InferenceResult>) {
        self.command_sender
            .send(RWKVCommand::Inference(params, sender))
            .unwrap();
    }

    // rwkv instance main loop
    pub fn spawn(
        &self,
        mode_path: String,
        tokenizer_path: String,
        n_threads: u32,
        load_result_sender: Sender<bool>,
        enable_logger: bool,
    ) {
        let rv = self.command_receiver.clone();

        thread::spawn(move || {
            let mut rwkv = RWKVInternal {
                context: RWKVContext::new(&mode_path, &tokenizer_path, n_threads),
            };

            if enable_logger {
                rwkv.context.rwkv_print_system_info_string();
            }

            load_result_sender.send(true).unwrap();

            let rv = rv.lock().unwrap();

            'rwkv_loop: loop {
                let command = rv.try_recv();
                match command {
                    Ok(RWKVCommand::Inference(params, sender)) => {
                        rwkv.inference(&params, &sender);
                    }
                    // Ok(RWKVCommand::Embedding(params, sender)) => {
                    //     rwkv.embedding(&params, &sender);
                    // }
                    Ok(RWKVCommand::Tokenize(text, sender)) => {
                        rwkv.tokenize(&text, &sender);
                    }
                    Err(TryRecvError::Disconnected) => {
                        break 'rwkv_loop;
                    }
                    _ => {
                        thread::yield_now();
                    }
                }
            }
        });
    }
}
